@@ -2,6 +2,7 @@ import asyncio
 import functools
 import inspect
 import time
+from sys import version_info as sys_version_info
 
 import pytest
 
@@ -44,16 +45,19 @@ def pytest_runtest_makereport(item, call):
         if hasattr(item, "start") and hasattr(item, "stop"):
             call.start = item.start
             call.stop = item.stop
+            call.duration = call.stop - call.start
 
     elif call.when == "setup":
         if hasattr(item, "start_setup") and hasattr(item, "stop_setup"):
             call.start = item.start_setup
             call.stop = item.stop_setup
+            call.duration = call.stop - call.start
 
     elif call.when == "teardown":
         if hasattr(item, "start_teardown") and hasattr(item, "stop_teardown"):
             call.start = item.start_teardown
             call.stop = item.stop_teardown
+            call.duration = call.stop - call.start
 
 
 def not_coroutine_failure(function_name: str, *args, **kwargs):
@@ -209,6 +213,12 @@ def pytest_runtestloop(session):
         else:
             regular_items.append(item)
 
+    def get_coro(task):
+        if sys_version_info >= (3, 8):
+            return task.get_coro()
+        else:
+            return task._coro
+
     async def run_tests(tasks, max_tasks):
         sidelined_tasks = tasks[max_tasks:]
         tasks = tasks[:max_tasks]
@@ -218,30 +228,44 @@ def pytest_runtestloop(session):
         completed = []
         while tasks:
 
+            # Schedule all the coroutines
+            for i in range(len(tasks)):
+                if asyncio.iscoroutine(tasks[i]):
+                    tasks[i] = asyncio.create_task(tasks[i])
+
             # Mark when the task was started
+            earliest_enqueue_time = time.time()
             for task in tasks:
                 if isinstance(task, asyncio.Task):
-                    item = item_by_coro[task._coro]
+                    item = item_by_coro[get_coro(task)]
                 else:
                     item = item_by_coro[task]
                 if not hasattr(item, "enqueue_time"):
                     item.enqueue_time = time.time()
+                earliest_enqueue_time = min(item.enqueue_time, earliest_enqueue_time)
 
+            time_to_wait = (time.time() - earliest_enqueue_time) - task_timeout
             done, pending = await asyncio.wait(
-                tasks, return_when=asyncio.FIRST_COMPLETED, timeout=30
+                tasks, return_when=asyncio.FIRST_COMPLETED, timeout=min(30, int(time_to_wait))
             )
 
             # Cancel tasks that have taken too long
             tasks = []
             for task in pending:
                 now = time.time()
-                item = item_by_coro[task._coro]
+                item = item_by_coro[get_coro(task)]
                 if task_timeout < now - item.enqueue_time:
-                    task.cancel()
+                    if sys_version_info >= (3, 9):
+                        msg = "Test took too long ({:.2f} s)".format(
+                            now - item.enqueue_time
+                        )
+                        task.cancel(msg=msg)
+                    else:
+                        task.cancel()
                 tasks.append(task)
 
             for result in done:
-                item = item_by_coro[result._coro]
+                item = item_by_coro[get_coro(result)]
 
                 # Flakey tests will be run again if they failed
                 # TODO: add retry count
